from cubicweb.schema import META_RTYPES

with cnx.deny_all_hooks_but("metadata"):
    scheme = cnx.find("ConceptScheme", title="SEDA : Niveaux de description").one()
    for concept in scheme.reverse_in_scheme:
        seda_label = concept.labels["seda"]
        seda2_label = {"recordgrp": "RecordGrp", "subgrp": "SubGrp"}.get(
            seda_label, seda_label.capitalize()
        )
        cnx.create_entity(
            "Label",
            label_of=concept,
            kind="preferred",
            language_code="seda-2",
            label=seda2_label,
        )
    commit()

sync_schema_props_perms(("SEDAArchiveTransfer", "title", "String"))

with cnx.deny_all_hooks_but():
    rql(
        'SET X user_annotation XI + "\n" + XUA WHERE X id XI, X user_annotation XUA, '
        "NOT X user_annotation NULL"
    )
    rql("SET X user_annotation XI WHERE X id XI, " "X user_annotation NULL")
    commit()

for etype in ("SEDAArchiveUnit", "SEDABinaryDataObject", "SEDAPhysicalDataObject"):
    drop_attribute(etype, "id")
    sync_schema_props_perms((etype, "user_annotation", "String"))

SKIP_RTYPES = {"seda_content", "container"} | META_RTYPES
for rschema, targets, role in schema.eschema("SEDAContent").relation_definitions():
    if rschema.type in SKIP_RTYPES:
        continue
    for target in targets:
        if role == "subject":
            add_relation_definition(
                "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement",
                rschema.type,
                target,
                ask_confirm=False,
            )
        else:
            add_relation_definition(
                target,
                rschema.type,
                "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement",
                ask_confirm=False,
            )
    with cnx.deny_all_hooks_but():
        if role == "subject":
            rql(
                "SET SEQ {0} X WHERE C seda_content SEQ, C {0} X".format(rschema),
                ask_confirm=False,
            )
        else:
            rql(
                "SET X {0} SEQ WHERE C seda_content SEQ, X {0} C".format(rschema),
                ask_confirm=False,
            )
        commit(ask_confirm=False)
drop_entity_type("SEDAContent")

add_cube("eac")
if "Agent" not in fsschema:
    drop_entity_type("Agent")

add_attribute("SEDAKeyword", "keyword_content")
with cnx.deny_all_hooks_but():
    rql(
        "SET KW keyword_content C WHERE KWC seda_keyword_content KW, KWC keyword_content C"
    )
commit()
drop_entity_type("SEDAKeywordContent")



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

from io import TextIOWrapper

try:
    import unittest2 as unittest
except ImportError:
    import unittest

from cubicweb import devtools, Binary  # noqa

from cubicweb_seda.xsd2yams import xsy_mapping, YamsSchemaGenerator, XSDMMapping


def readable_edef(edef):
    def name(e):
        return getattr(e, "local_name", e.__class__.__name__)

    if len(edef) == 3:
        return (
            name(edef[0]),
            edef[1],
            [(name(rdef[0].target), [x[:-1] for x in rdef[1]]) for rdef in edef[2]],
        )
    else:
        return (
            edef[0],
            [(name(rdef[0].target), [x[:-1] for x in rdef[1]]) for rdef in edef[1]],
        )


class CodeGenerationTC(unittest.TestCase):
    def test_base(self):
        mapping = xsy_mapping("DataObjectPackage")
        stream = TextIOWrapper(Binary())
        YamsSchemaGenerator().generate(mapping, stream)
        code = stream.read()
        # assert this is valid python
        compile(code, "<generated schema>", mode="exec")
        # assert there is no duplicated classes
        classes = set()
        for line in code.splitlines():
            if line.startswith("class "):
                # don't consider case, postgres won't
                classname = line.split()[1].split("(:")[0].lower()
                self.assertNotIn(classname, classes)
                classes.add(classname)
            elif line.strip().startswith("name ="):
                name = line.split("name =")[-1].strip()[1:-1]
                self.assertLess(len(name), 64, f"too long name {name!r} ({len(name)}))")


class XSYMappingTC(unittest.TestCase):
    def assertEMappingEqual(self, emapping, attributes={}):
        self.assertEqual(emapping.attributes, attributes)

    def assertRMappingEqual(self, rmapping, rdefs):
        self.assertEqual(
            sorted((rdef.subjtype, rdef.objtype) for rdef in rmapping.rdefs), rdefs
        )

    def test_keyword(self):
        mapping = xsy_mapping("Keyword")
        emapping = mapping.etypes["SEDAKeyword"]
        self.assertEMappingEqual(emapping, {"keyword_content": "String"})

    def test_binarydataobject(self):
        mapping = xsy_mapping("BinaryDataObject")
        self.assertIn("SEDABinaryDataObject", mapping.etypes)
        emapping = mapping.etypes["SEDABinaryDataObject"]
        self.assertEMappingEqual(emapping, {"id": "String", "filename": "String"})
        emapping = mapping.etypes["SEDACompressed"]
        self.assertEMappingEqual(emapping, {"compressed": "String"})

    def test_hardcoded_cardinalities(self):
        mapping = xsy_mapping("Content")
        seda_title_card = [
            rdefmapping.card
            for rdefmapping in mapping.ordered
            if getattr(rdefmapping, "rtype", None) == "seda_title"
        ][0]
        self.assertEqual(seda_title_card, "11")
        seda_description_card = [
            rdefmapping.card
            for rdefmapping in mapping.ordered
            if getattr(rdefmapping, "rtype", None) == "seda_description"
        ][0]
        self.assertEqual(seda_description_card, "1?")


class XSIterateTC(unittest.TestCase):
    def test_custodialhistoryitem(self):
        element_defs = iter(XSDMMapping("CustodialHistoryItem"))
        edef = next(element_defs)
        self.assertEqual(
            readable_edef(edef),
            (
                "CustodialHistoryItem",
                "SEDACustodialHistoryItem",
                [
                    (
                        "when",
                        [
                            ("seda_when", "object", "SEDAwhen"),
                            ("when", "subject", "TZDatetime"),
                        ],
                    ),
                    (
                        "CustodialHistoryItem",
                        [("custodial_history_item", "subject", "String")],
                    ),
                ],
            ),
        )

    def test_keyword(self):
        element_defs = iter(XSDMMapping("Keyword"))
        edef = next(element_defs)
        self.assertEqual(
            readable_edef(edef),
            (
                "Keyword",
                "SEDAKeyword",
                [
                    (
                        "id",
                        [("seda_id", "object", "SEDAid"), ("id", "subject", "String")],
                    ),
                    ("KeywordContent", []),
                    (
                        "KeywordReference",
                        [
                            (
                                "seda_keyword_reference_from",
                                "object",
                                "SEDAKeywordReference",
                            ),
                            ("seda_keyword_reference_to", "subject", "Concept"),
                        ],
                    ),
                    (
                        "KeywordType",
                        [
                            ("seda_keyword_type_from", "object", "SEDAKeywordType"),
                            ("seda_keyword_type_to", "subject", "Concept"),
                        ],
                    ),
                ],
            ),
        )
        edef = next(element_defs)
        self.assertEqual(
            readable_edef(edef),
            (
                "KeywordContent",
                "SEDAKeyword",
                [
                    ("KeywordContent", [("keyword_content", "subject", "String")]),
                ],
            ),
        )
        self.assertRaises(StopIteration, next, element_defs)

    def test_rule(self):
        element_defs = iter(XSDMMapping("AppraisalRule"))
        edef = next(element_defs)
        self.assertEqual(
            readable_edef(edef),
            (
                "AppraisalRule",
                "SEDAAppraisalRule",
                [
                    (
                        "Sequence",
                        [
                            (
                                "seda_seq_appraisal_rule_rule",
                                "subject",
                                "SEDASeqAppraisalRuleRule",
                            )
                        ],
                    ),
                    (
                        "Alternative",
                        [
                            (
                                "seda_alt_appraisal_rule_prevent_inheritance",
                                "subject",
                                "SEDAAltAppraisalRulePreventInheritance",
                            )
                        ],
                    ),
                    ("FinalAction", [("seda_final_action", "subject", "Concept")]),
                ],
            ),
        )
        edef = next(element_defs)
        self.assertEqual(
            readable_edef(edef),
            (
                "Sequence",
                "SEDASeqAppraisalRuleRule",
                [
                    ("Rule", [("seda_rule", "subject", "Concept")]),
                    (
                        "StartDate",
                        [
                            ("seda_start_date", "object", "SEDAStartDate"),
                            ("start_date", "subject", "Date"),
                        ],
                    ),
                ],
            ),
        )
        edef = next(element_defs)
        self.assertEqual(
            readable_edef(edef),
            (
                "Alternative",
                "SEDAAltAppraisalRulePreventInheritance",
                [
                    (
                        "PreventInheritance",
                        [
                            (
                                "seda_prevent_inheritance",
                                "object",
                                "SEDAPreventInheritance",
                            ),
                            ("prevent_inheritance", "subject", "Boolean"),
                        ],
                    ),
                    (
                        "RefNonRuleId",
                        [
                            ("seda_ref_non_rule_id_from", "object", "SEDARefNonRuleId"),
                            ("seda_ref_non_rule_id_to", "subject", "Concept"),
                        ],
                    ),
                ],
            ),
        )
        edef = next(element_defs)
        self.assertEqual(readable_edef(edef), ("StartDate", "SEDAStartDate", []))
        edef = next(element_defs)
        self.assertEqual(
            readable_edef(edef), ("PreventInheritance", "SEDAPreventInheritance", [])
        )
        self.assertRaises(StopIteration, next, element_defs)

    def test_binarydataobject(self):
        element_defs = XSDMMapping("BinaryDataObject")
        edef = next(iter(element_defs))
        children = [
            getattr(e[0].target, "local_name", e[0].target.__class__.__name__)
            for e in edef[2]
        ]
        self.assertEqual(
            children,
            [
                "id",
                "DataObjectSystemId",
                "DataObjectGroupSystemId",
                "Relationship",
                "DataObjectVersion",
                "Alternative",
                "MessageDigest",
                "Size",
                "Compressed",
                "FormatIdentification",
                "FileInfo",
            ],
        )
        digest_element = edef[2][6][0].target
        self.assertEqual(edef[2][6][1], [])
        digest_edef = (digest_element,) + element_defs[digest_element]
        self.assertEqual(
            readable_edef(digest_edef),
            (
                "MessageDigest",
                "SEDABinaryDataObject",
                [
                    ("algorithm", [("seda_algorithm", "subject", "Concept")]),
                ],
            ),
        )

    def test_archiveunit(self):
        element_defs = XSDMMapping("ArchiveUnit")
        edef = next(iter(element_defs))
        self.assertEqual(
            readable_edef(edef),
            (
                "ArchiveUnit",
                "SEDAArchiveUnit",
                [
                    ("id", [("id", "subject", "String")]),
                    (
                        "Alternative",
                        [
                            (
                                "seda_alt_archive_unit_archive_unit_ref_id",
                                "subject",
                                "SEDAAltArchiveUnitArchiveUnitRefId",
                            )
                        ],
                    ),
                ],
            ),
        )
        alt_element = edef[2][-1][0].target
        alt_edef = (alt_element,) + element_defs[alt_element]
        self.assertEqual(
            readable_edef(alt_edef),
            (
                "Alternative",
                "SEDAAltArchiveUnitArchiveUnitRefId",
                [
                    (
                        "ArchiveUnitRefId",
                        [
                            (
                                "seda_archive_unit_ref_id_from",
                                "object",
                                "SEDAArchiveUnitRefId",
                            ),
                            (
                                "seda_archive_unit_ref_id_to",
                                "subject",
                                "SEDAArchiveUnit",
                            ),
                        ],
                    ),
                    (
                        "Sequence",
                        [
                            (
                                "seda_seq_alt_archive_unit_archive_unit_ref_id_management",
                                "subject",
                                "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement",
                            )
                        ],
                    ),
                ],
            ),
        )
        seq_element = alt_edef[2][-1][0].target
        seq_edef = (seq_element,) + element_defs[seq_element]
        self.assertEqual(
            readable_edef(seq_edef),
            (
                "Sequence",
                "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement",
                [
                    ("Management", []),
                    ("Content", []),
                    (
                        "ArchiveUnit",
                        [("seda_archive_unit", "object", "SEDAArchiveUnit")],
                    ),
                    (
                        "DataObjectReference",
                        [
                            (
                                "seda_data_object_reference",
                                "object",
                                "SEDADataObjectReference",
                            )
                        ],
                    ),
                ],
            ),
        )

    def test_dataobjectpackage(self):
        element_defs = XSDMMapping("ArchiveTransfer")
        edef = next(iter(element_defs))
        self.assertEqual(
            readable_edef(edef),
            (
                "ArchiveTransfer",
                "SEDAArchiveTransfer",
                [
                    (
                        "id",
                        [("seda_id", "object", "SEDAid"), ("id", "subject", "String")],
                    ),
                    ("Comment", [("seda_comment", "object", "SEDAComment")]),
                    ("Date", [("date", "subject", "TZDatetime")]),
                    (
                        "MessageIdentifier",
                        [("message_identifier", "subject", "String")],
                    ),
                    ("Signature", [("seda_signature", "object", "SEDASignature")]),
                    (
                        "ArchivalAgreement",
                        [
                            (
                                "seda_archival_agreement",
                                "object",
                                "SEDAArchivalAgreement",
                            ),
                            ("archival_agreement", "subject", "String"),
                        ],
                    ),
                    ("CodeListVersions", []),
                    ("DataObjectPackage", []),
                    (
                        "RelatedTransferReference",
                        [
                            (
                                "seda_related_transfer_reference",
                                "object",
                                "SEDARelatedTransferReference",
                            ),
                            ("related_transfer_reference", "subject", "String"),
                        ],
                    ),
                    (
                        "TransferRequestReplyIdentifier",
                        [
                            (
                                "seda_transfer_request_reply_identifier",
                                "object",
                                "SEDATransferRequestReplyIdentifier",
                            ),
                            ("transfer_request_reply_identifier", "subject", "String"),
                        ],
                    ),
                    (
                        "ArchivalAgency",
                        [("seda_archival_agency", "subject", "AuthorityRecord")],
                    ),
                    (
                        "TransferringAgency",
                        [("seda_transferring_agency", "subject", "AuthorityRecord")],
                    ),
                ],
            ),
        )
        dao_element = edef[2][7][0].target
        dao_edef = (dao_element,) + element_defs[dao_element]
        descr_element = dao_edef[2][4][0].target
        descr_edef = (descr_element,) + element_defs[descr_element]
        self.assertEqual(
            readable_edef(descr_edef),
            (
                "DescriptiveMetadata",
                "SEDAArchiveTransfer",
                [("ArchiveUnit", [("seda_archive_unit", "object", "SEDAArchiveUnit")])],
            ),
        )
        mgmt_element = dao_edef[2][5][0].target
        mgmt_edef = (mgmt_element,) + element_defs[mgmt_element]
        self.assertEqual(
            readable_edef(mgmt_edef),
            (
                "ManagementMetadata",
                "SEDAArchiveTransfer",
                [
                    (
                        "id",
                        [("seda_id", "object", "SEDAid"), ("id", "subject", "String")],
                    ),
                    (
                        "ArchivalProfile",
                        [
                            ("seda_archival_profile", "object", "SEDAArchivalProfile"),
                            ("archival_profile", "subject", "String"),
                        ],
                    ),
                    (
                        "ServiceLevel",
                        [
                            ("seda_service_level", "object", "SEDAServiceLevel"),
                            ("service_level", "subject", "String"),
                        ],
                    ),
                    (
                        "AcquisitionInformation",
                        [
                            (
                                "seda_acquisition_information_from",
                                "object",
                                "SEDAAcquisitionInformation",
                            ),
                            ("seda_acquisition_information_to", "subject", "Concept"),
                        ],
                    ),
                    (
                        "LegalStatus",
                        [
                            ("seda_legal_status_from", "object", "SEDALegalStatus"),
                            ("seda_legal_status_to", "subject", "Concept"),
                        ],
                    ),
                    (
                        "OriginatingAgencyIdentifier",
                        [
                            (
                                "seda_originating_agency_identifier",
                                "object",
                                "SEDAOriginatingAgencyIdentifier",
                            ),
                            ("originating_agency_identifier", "subject", "String"),
                        ],
                    ),
                    (
                        "SubmissionAgencyIdentifier",
                        [
                            (
                                "seda_submission_agency_identifier",
                                "object",
                                "SEDASubmissionAgencyIdentifier",
                            ),
                            ("submission_agency_identifier", "subject", "String"),
                        ],
                    ),
                    (
                        "StorageRule",
                        [("seda_storage_rule", "object", "SEDAStorageRule")],
                    ),
                    (
                        "AppraisalRule",
                        [("seda_appraisal_rule", "object", "SEDAAppraisalRule")],
                    ),
                    ("AccessRule", [("seda_access_rule", "object", "SEDAAccessRule")]),
                    (
                        "DisseminationRule",
                        [
                            (
                                "seda_dissemination_rule",
                                "object",
                                "SEDADisseminationRule",
                            )
                        ],
                    ),
                    ("ReuseRule", [("seda_reuse_rule", "object", "SEDAReuseRule")]),
                    (
                        "ClassificationRule",
                        [
                            (
                                "seda_classification_rule",
                                "object",
                                "SEDAClassificationRule",
                            )
                        ],
                    ),
                    ("LogBook", [("seda_log_book", "object", "SEDALogBook")]),
                    (
                        "NeedAuthorization",
                        [
                            (
                                "seda_need_authorization",
                                "object",
                                "SEDANeedAuthorization",
                            ),
                            ("need_authorization", "subject", "Boolean"),
                        ],
                    ),
                ],
            ),
        )

    def test_elements_by_name(self):
        mapping = XSDMMapping("ArchiveTransfer")
        self.assertEqual(len(mapping.elements_by_name("StartDate")), 2)
        self.assertEqual(len(mapping.elements_by_name("CodeListVersions")), 1)
        self.assertEqual(
            mapping.element_by_name("CodeListVersions").local_name, "CodeListVersions"
        )


if __name__ == "__main__":
    unittest.main()



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""Generate CubicWeb's entities from XSD file"""

import json

from cubicweb_seda.xsd2yams import MULTIPLE_CHILDREN, CodeGenerator
from cubicweb_seda.xsd2uicfg import ordered_attributes

MULTIPLE_ETYPES = {etype for etype, _ in MULTIPLE_CHILDREN}

TEMPLATE = """class {etype}(SEDAAnyEntity):
    __regid__ = '{etype}'
    fetch_attrs, cw_fetch_order = fetch_config({attributes})
    value_attr = {value_attribute}
"""


class EntitiesGenerator(CodeGenerator):
    def _generate(self, mapping, stream):
        stream.write("from cubicweb.entities import AnyEntity, fetch_config\n\n")
        stream.write(
            """
class SEDAAnyEntity(AnyEntity):
    __abstract__ = True
    value_attr = None

    def dc_title(self):
        if self.value_attr is None:
            return self.dc_type()
        return self.printable_value(self.value_attr)


"""
        )
        self.altetype2rtype = {}  # rtype to children elements for all choices
        self.check_card_etypes = set()  # entity types whose cardinality must be checked
        self.check_card_rtypes = (
            set()
        )  # relation types whose children cardinality must be checked
        for mapping_element in mapping.ordered:
            self._callback("register_alt_rtype_for", mapping_element)
            eclass_code = self._callback("entity_class_for", mapping_element)
            if eclass_code:
                stream.write(eclass_code + "\n")
        # Write pre-computed data structures as constants
        for name, data in [
            ("CHOICE_RTYPE", self.altetype2rtype),
            ("CHECK_CARD_ETYPES", sorted(self.check_card_etypes)),
            ("CHECK_CHILDREN_CARD_RTYPES", sorted(self.check_card_rtypes)),
        ]:
            stream.write(f"{name} = ")
            stream.write(
                json.dumps(
                    data, sort_keys=True, indent=2, separators=(",", ": ")
                ).replace('"', "'")
                + "\n"
            )

    def entity_class_for_e_type_mapping(self, mapping):
        attributes = ordered_attributes(mapping)
        value_attribute = None
        if mapping.etype in (
            "SEDAArchiveUnit",
            "SEDABinaryDataObject",
            "SEDAPhysicalDataObject",
            "SEDADataObjectGroup",
        ):
            pass
        elif mapping.attributes:
            value_attribute = next(iter(mapping.attributes))
        elif mapping.cards and len(mapping.cards) > 1:
            attributes = ["user_cardinality"]
        if mapping.etype in MULTIPLE_ETYPES:
            attributes.insert(0, "ordering")
        return TEMPLATE.format(
            etype=mapping.etype,
            attributes=attributes,
            value_attribute=repr(value_attribute),
        )

    def register_alt_rtype_for_rdef_mapping(self, mapping):
        composite = mapping.composite
        if not composite:
            return
        altetypes = mapping.objtypes if composite == "object" else [mapping.subjtype]
        for altetype in altetypes:
            if altetype.startswith("SEDAAlt"):
                self.altetype2rtype.setdefault(altetype, []).append(
                    (mapping.rtype, composite)
                )

        if mapping.card[1] in "*+":
            self.check_card_etypes.add(mapping.subjtype)
            self.check_card_rtypes.add(mapping.rtype)


if __name__ == "__main__":
    EntitiesGenerator.main()



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda unit tests for schema"""

import uuid

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.schema import ERQLExpression, RRQLExpression

from cubicweb_seda import iter_all_rdefs

from cubicweb_seda import testutils


class SchemaConceptConstraintsTC(CubicWebTC):
    assertValidationError = testutils.assertValidationError

    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            mt_scheme = testutils.scheme_for_type(cnx, "file_category", None)
            mt_concept = mt_scheme.add_concept(label="text/plain")
            enc_scheme = testutils.scheme_for_type(cnx, "seda_encoding_to", None)
            enc_concept = enc_scheme.add_concept(label="utf-8")
            cnx.commit()
            self.mt_scheme = mt_scheme.eid
            self.mt_concept = mt_concept.eid
            self.enc_scheme = enc_scheme.eid
            self.enc_concept = enc_concept.eid

    def test_code_lists_constraints_simple(self):
        with self.admin_access.cnx() as cnx:
            bdo = testutils.create_transfer_to_bdo(cnx)
            transfer = bdo.container[0]
            cnx.create_entity(
                "SEDAMessageDigestAlgorithmCodeListVersion",
                seda_message_digest_algorithm_code_list_version_from=transfer,
                seda_message_digest_algorithm_code_list_version_to=self.enc_scheme,
            )
            cnx.commit()

            with self.assertValidationError(cnx) as cm:
                bdo.cw_set(seda_algorithm=self.mt_concept)
            self.assertIn("seda_algorithm-subject", cm.exception.errors)

            bdo.cw_set(seda_algorithm=self.enc_concept)
            cnx.commit()

    def assertMimeTypeConcept(self, bdo):
        cnx = bdo._cw

        with self.assertValidationError(cnx) as cm:
            bdo.mime_type.cw_set(seda_mime_type_to=self.enc_concept)
        self.assertIn("seda_mime_type_to-subject", cm.exception.errors)

        bdo.mime_type.cw_set(seda_mime_type_to=self.mt_concept)
        cnx.commit()

    def test_archive_transfer_mime_type_constraint(self):
        with self.admin_access.cnx() as cnx:
            bdo = testutils.create_transfer_to_bdo(cnx)
            cnx.commit()  # will setup code list version
            self.assertMimeTypeConcept(bdo)

    def test_component_archive_unit_mime_type_constraint(self):
        with self.admin_access.cnx() as cnx:
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(None, cnx=cnx)
            bdo = testutils.create_data_object(unit_alt_seq)
            cnx.commit()
            self.assertMimeTypeConcept(bdo)

    def assertEncodingConcept(self, bdo):
        cnx = bdo._cw

        with self.assertValidationError(cnx) as cm:
            cnx.create_entity(
                "SEDAEncoding", seda_encoding_from=bdo, seda_encoding_to=self.mt_concept
            )
        self.assertIn("seda_encoding_to-subject", cm.exception.errors)

        cnx.create_entity(
            "SEDAEncoding", seda_encoding_from=bdo, seda_encoding_to=self.enc_concept
        )
        cnx.commit()

    def test_archive_transfer_encoding_constraint(self):
        with self.admin_access.cnx() as cnx:
            bdo = testutils.create_transfer_to_bdo(cnx)
            cnx.commit()  # will setup code list version
            self.assertEncodingConcept(bdo)

    def test_component_archive_unit_encoding_constraint(self):
        with self.admin_access.cnx() as cnx:
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(None, cnx=cnx)
            bdo = testutils.create_data_object(unit_alt_seq)
            cnx.commit()
            self.assertEncodingConcept(bdo)

    def assertDigestAlgorithmConcept(self, bdo):
        cnx = bdo._cw

        with self.assertValidationError(cnx) as cm:
            bdo.cw_set(seda_algorithm=self.mt_concept)
        self.assertIn("seda_algorithm-subject", cm.exception.errors)

        bdo.cw_set(seda_algorithm=self.enc_concept)
        cnx.commit()

    def test_archive_transfer_digest_algorithm_constraint(self):
        with self.admin_access.cnx() as cnx:
            bdo = testutils.create_transfer_to_bdo(cnx)
            cnx.commit()  # commit first to get the container
            transfer = bdo.container[0]
            cnx.create_entity(
                "SEDAMessageDigestAlgorithmCodeListVersion",
                seda_message_digest_algorithm_code_list_version_from=transfer,
                seda_message_digest_algorithm_code_list_version_to=self.enc_scheme,
            )
            cnx.commit()
            self.assertDigestAlgorithmConcept(bdo)

    def test_component_archive_unit_digest_algorithm_constraint(self):
        with self.admin_access.cnx() as cnx:
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(None, cnx=cnx)
            bdo = testutils.create_data_object(unit_alt_seq)
            bdo_type = cnx.find("CWEType", name="SEDABinaryDataObject").one()
            alg_type = cnx.find("CWRType", name="seda_algorithm").one()
            cnx.entity_from_eid(self.enc_scheme).cw_set(
                scheme_entity_type=bdo_type, scheme_relation_type=alg_type
            )
            cnx.commit()
            self.assertDigestAlgorithmConcept(bdo)


class SchemaTC(CubicWebTC):
    assertValidationError = testutils.assertValidationError

    def test_component_archive_unit_rule_constraint(self):
        with self.admin_access.cnx() as cnx:
            for rule_type in ("access", "appraisal"):
                etype = f"SEDASeq{rule_type.capitalize()}RuleRule"
                scheme = testutils.scheme_for_type(cnx, "seda_rule", etype)
                concept = scheme.add_concept(label="whatever")
                cnx.commit()

                unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(
                    None, cnx=cnx
                )
                rule_rule = cnx.create_entity(
                    etype.format(rule_type.capitalize()), seda_rule=concept
                )
                cnx.create_entity(
                    f"SEDA{rule_type.capitalize()}Rule",
                    **{
                        f"seda_{rule_type}_rule": unit_alt_seq,
                        f"seda_seq_{rule_type}_rule_rule": rule_rule,
                    },
                )
                cnx.commit()

    def test_rdef_container_permissions(self):
        """Check that permissions are correctly set on rdefs between entity types contained in
        SEDAArchiveTransfer."""
        for action in ("add", "delete"):
            # Parent is container
            rdef = self.schema["seda_comment"].rdef(
                "SEDAComment", "SEDAArchiveTransfer"
            )
            (rqlexpr,) = (
                p for p in rdef.permissions[action] if isinstance(p, RRQLExpression)
            )
            self.assertEqual(rqlexpr.expression, "U has_update_permission O")
            # Parent is contained in container
            rdef = self.schema["seda_tag"].rdef(
                "SEDATag", "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement"
            )
            (rqlexpr,) = (
                p for p in rdef.permissions[action] if isinstance(p, RRQLExpression)
            )
            self.assertEqual(
                rqlexpr.expression, "U has_update_permission C, O container C"
            )
            # Parent is subject
            rdef = self.schema[
                "seda_seq_alt_archive_unit_archive_unit_ref_id_management"
            ].rdef(
                "SEDAAltArchiveUnitArchiveUnitRefId",
                "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement",
            )
            (rqlexpr,) = (
                p for p in rdef.permissions[action] if isinstance(p, RRQLExpression)
            )
            self.assertEqual(
                rqlexpr.expression, "U has_update_permission C, S container C"
            )

    def test_etype_container_permissions(self):
        """Check that permissions are correctly set on entity types contained in
        SEDAArchiveTransfer."""
        for action in ("update", "delete"):
            # Parent is container
            etype = self.schema["SEDAComment"]
            (rqlexpr,) = (
                p for p in etype.permissions[action] if isinstance(p, ERQLExpression)
            )
            self.assertEqual(
                rqlexpr.expression,
                f"U has_{action}_permission C, X container C",
            )

    def test_archive_unit_permissions(self):
        """Check that permissions are correctly set on archive unit that may be either container
        or contained."""
        for action in ("update", "delete"):
            etype = self.schema["SEDAArchiveUnit"]
            rqlexpr1, rqlexpr2 = (
                p for p in etype.permissions[action] if isinstance(p, ERQLExpression)
            )
            self.assertEqual(
                rqlexpr1.expression,
                f"U has_{action}_permission C, X container C",
            )
            self.assertEqual(
                rqlexpr2.expression,
                "NOT EXISTS(X container C), U in_group G, "
                'G name IN("managers", "users")',
            )

    def test_rule_default_cardinality(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            for rule_type in ("access", "appraisal"):
                rule_etype = f"SEDA{rule_type.capitalize()}Rule"
                rule_rtype = f"seda_{rule_type}_rule"
                rule = cnx.create_entity(rule_etype, **{rule_rtype: transfer})
                rule_rule_etype = f"SEDASeq{rule_type.capitalize()}RuleRule"
                rule_rule_rtype = f"reverse_seda_seq_{rule_type}_rule_rule"
                rule_rule = cnx.create_entity(
                    rule_rule_etype, **{rule_rule_rtype: rule}
                )
                self.assertEqual(rule_rule.user_cardinality, "1")
                start_date = cnx.create_entity(
                    "SEDAStartDate", seda_start_date=rule_rule
                )
                self.assertEqual(start_date.user_cardinality, "1")

    def test_fti(self):
        # "Reverse" text to be searched in order not to be troubled by other
        # entities that may live in the DB (e.g. Concepts) with similar text.
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="Profile")
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(
                transfer, title="transfer name"[::-1]
            )
            cnx.create_entity(
                "SEDAAccessRule",
                seda_access_rule=unit_alt_seq,
                user_annotation="some annotation"[::-1],
            )
            testutils.create_data_object(transfer, filename="fixed.txt"[::-1])
            cnx.commit()

            for search in ("name", "annotation", "fixed"):
                with self.subTest(search=search):
                    rset = cnx.execute(
                        "Any X WHERE X has_text %(search)s", {"search": search[::-1]}
                    )
                    self.assertEqual([r for r, in rset.rows], [transfer.eid])

    def test_scheme_code_keyword_type_constraint(self):
        with self.admin_access.cnx() as cnx:
            ckt_scheme = testutils.scheme_for_type(cnx, "seda_keyword_type_to", None)
            ckt_concept = ckt_scheme.add_concept(label="geoname")
            scheme = testutils.scheme_for_type(cnx, "seda_mime_type_to", None)
            concept = scheme.add_concept(label="text/plain")
            cnx.commit()

            with self.assertValidationError(cnx):
                scheme.cw_set(code_keyword_type=concept)

            scheme.cw_set(code_keyword_type=ckt_concept)
            cnx.commit()

    def test_container_relation(self):
        container_rtype = self.schema["container"]
        for rdef in container_rtype.rdefs.values():
            if rdef.subject == "SEDAArchiveUnit":
                self.assertEqual(rdef.cardinality, "?*")
            else:
                self.assertEqual(rdef.cardinality, "1*")
            self.assertEqual(
                rdef.permissions,
                {"add": (), "delete": (), "read": ("managers", "users", "guests")},
            )

    def test_reference_to_data_object_is_not_composite(self):
        self._test_reference_to_data_object_composite(False)

    def test_simplified_reference_to_data_object_is_composite(self):
        self._test_reference_to_data_object_composite(True)

    def _test_reference_to_data_object_composite(self, simplified):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity(
                "SEDAArchiveTransfer",
                title="test profile",
                simplified_profile=simplified,
            )
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            bdo = testutils.create_data_object(
                unit_alt_seq, seda_binary_data_object=transfer
            )
            cnx.commit()
            reference = bdo.reverse_seda_data_object_reference_id[0]
            reference.cw_delete()
            cnx.commit()
            self.assertEqual(
                len(cnx.execute("Any X WHERE X eid %(x)s", {"x": bdo.eid})),
                0 if simplified else 1,
            )

    def test_delete_archiveunit_with_data_object(self):
        """Make sure deletion of an archive unit with a binary data object,
        but not related to an archive transfer is possible.
        """
        with self.admin_access.cnx() as cnx:
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(None, cnx=cnx)
            bdo = testutils.create_data_object(unit_alt_seq)
            cnx.commit()
            bdo.cw_clear_all_caches()
            unit.cw_delete()
            cnx.commit()
            self.assertFalse(cnx.execute("Any X WHERE X eid %(x)s", {"x": bdo.eid}))


class SecurityTC(CubicWebTC):
    assertUnauthorized = testutils.assertUnauthorized

    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            self.create_user(cnx, login="alice")
            cnx.commit()

    def test_profile(self):
        with self.admin_access.cnx() as cnx:
            self.create_user(cnx, login="bob")
            cnx.commit()
        with self.new_access("alice").cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="Alice Profile")
            testutils.create_archive_unit(transfer)
            testutils.create_authority_record(
                cnx,
                "Archival inc.",
                reverse_seda_archival_agency=transfer,
                record_id=str(uuid.uuid1()),
            )
            cnx.create_entity("SEDAComment", comment="Whooot.", seda_comment=transfer)
            scheme = cnx.create_entity("ConceptScheme", title="Some vocabulary")
            cnx.create_entity(
                "SEDAMimeTypeCodeListVersion",
                seda_mime_type_code_list_version_from=transfer,
                seda_mime_type_code_list_version_to=scheme,
            )
            cnx.commit()
        with self.new_access("bob").cnx() as cnx:
            transfer = cnx.entity_from_eid(transfer.eid)
            # modification of the container
            with self.assertUnauthorized(cnx):
                transfer.cw_set(title="Bob Profile")
            # modification of a contained entity
            comment = transfer.reverse_seda_comment[0]
            with self.assertUnauthorized(cnx):
                comment.cw_set(comment="You got hacked")
            with self.assertUnauthorized(cnx):
                cnx.create_entity(
                    "SEDAArchivalAgreement", seda_archival_agreement=transfer
                )
            # modification of a relation from the container to a non contained entity
            with self.assertUnauthorized(cnx):
                testutils.create_authority_record(
                    cnx,
                    name="Bob Archival inc.",
                    reverse_seda_archival_agency=transfer,
                    record_id=str(uuid.uuid1()),
                )
            # modification of a relation from a contained entity to a non contained entity
            mtclv = transfer.reverse_seda_mime_type_code_list_version_from[0]
            with self.assertUnauthorized(cnx):
                scheme = cnx.create_entity(
                    "ConceptScheme", title="Some nasty vocabulary"
                )
                mtclv.cw_set(seda_mime_type_code_list_version_to=scheme)
            # deletion of a contained entity
            with self.assertUnauthorized(cnx):
                comment.cw_delete()
            # deletion of a outer relation
            with self.assertUnauthorized(cnx):
                transfer.reverse_seda_mime_type_code_list_version_from[0].cw_set(
                    seda_mime_type_code_list_version_to=None
                )
            # deletion of an archive unit
            with self.assertUnauthorized(cnx):
                transfer.archive_units[0].cw_delete()
            # deletion of the container
            with self.assertUnauthorized(cnx):
                transfer.cw_delete()

        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(transfer.eid)
            # ensure every subobjects permissions depends on top-level
            # permissions (don't even include managers group)
            with self.temporary_permissions(
                (self.schema["SEDAArchiveTransfer"], {"update": (), "delete": ()})
            ):
                # modification of a contained entity
                comment = transfer.reverse_seda_comment[0]
                with self.assertUnauthorized(cnx):
                    comment.cw_set(comment="You got hacked")
                with self.assertUnauthorized(cnx):
                    comment.cw_delete()
                with self.assertUnauthorized(cnx):
                    cnx.create_entity(
                        "SEDAArchivalAgreement", seda_archival_agreement=transfer
                    )
                # modification of a relation from the container to a non contained entity
                with self.assertUnauthorized(cnx):
                    testutils.create_authority_record(
                        cnx,
                        name="Bob Archival inc.",
                        reverse_seda_archival_agency=transfer,
                        record_id=str(uuid.uuid1()),
                    )
                # deletion of an archive unit
                with self.assertUnauthorized(cnx):
                    transfer.archive_units[0].cw_delete()
                # deletion of the container
                with self.assertUnauthorized(cnx):
                    transfer.cw_delete()

    def test_archive_unit(self):
        with self.admin_access.cnx() as cnx:
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(None, cnx=cnx)
            cnx.commit()

            # unit has no parent, modifications are allowed.
            unit.cw_set(user_annotation="argh")
            unit_alt_seq.reverse_seda_title[0].cw_set(title="gloup")
            cnx.commit()

        with self.new_access("anon").cnx() as cnx:
            title = cnx.entity_from_eid(unit_alt_seq.reverse_seda_title[0].eid)
            unit = cnx.entity_from_eid(unit.eid)
            with self.assertUnauthorized(cnx):
                title.cw_set(title="zorglub")
            with self.assertUnauthorized(cnx):
                unit.cw_set(user_annotation="zorglub")
            with self.assertUnauthorized(cnx):
                cnx.create_entity("SEDADescription", seda_description=unit_alt_seq)
            with self.assertUnauthorized(cnx):
                title.cw_delete()
            with self.assertUnauthorized(cnx):
                unit.cw_delete()
        with self.admin_access.cnx() as cnx:
            unit = cnx.entity_from_eid(unit.eid)
            unit.cw_delete()
            cnx.commit()

    def test_users_can_create_unit(self):
        with self.new_access("alice").cnx() as cnx:
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(None, cnx=cnx)
            cnx.commit()

    def test_users_can_clone(self):
        """Functional test for SEDA component clone."""
        with self.admin_access.cnx() as cnx:
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(None, cnx=cnx)
            cnx.commit()

        with self.new_access("alice").cnx() as cnx:
            unit = cnx.entity_from_eid(unit.eid)
            cloned = cnx.create_entity(
                unit.cw_etype, user_annotation="x", clone_of=unit
            )
            cnx.commit()
            cloned.cw_clear_all_caches()

            self.assertTrue(cloned.first_level_choice.content_sequence)

    def test_scheme_type(self):
        with self.new_access("alice").cnx() as cnx:
            scheme = cnx.create_entity("ConceptScheme", title="Some vocabulary")
            cnx.commit()

            etype = cnx.find("CWEType", name="CWEType").one()
            with self.assertUnauthorized(cnx):
                scheme.cw_set(scheme_entity_type=etype)

            rtype = cnx.find("CWRType", name="name").one()
            with self.assertUnauthorized(cnx):
                scheme.cw_set(scheme_relation_type=rtype)

    def test_scheme_code_keyword_type(self):
        with self.admin_access.cnx() as cnx:
            admin_scheme = testutils.scheme_for_type(
                cnx, "seda_keyword_type_to", None, "type1"
            )
            cnx.commit()
            type_concept = admin_scheme.reverse_in_scheme[0]

        with self.new_access("alice").cnx() as cnx:
            # ensure code_keyword_type is editable by standard user
            scheme = cnx.create_entity(
                "ConceptScheme", title="Some vocabulary", code_keyword_type=type_concept
            )
            cnx.commit()
            scheme.cw_set(code_keyword_type=None)
            cnx.commit()


class SchemaIteratorTC(CubicWebTC):
    def test_iter_all_rdefs(self):
        values = {
            str(rdef)
            for rdef, role in iter_all_rdefs(self.schema, "SEDABinaryDataObject")
        }
        # composite relation
        self.assertIn(
            "relation SEDACreatingOs seda_creating_os SEDABinaryDataObject", values
        )
        # external relation
        self.assertIn(
            "relation SEDARelationship seda_type_relationship Concept", values
        )
        # outside this compound graph
        self.assertNotIn(
            "relation SEDADataObjectVersion seda_data_object_version_from SEDAPhysicalDataObject",
            values,
        )


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda data import tools"""

from itertools import count
from os.path import abspath, dirname, join

from cubicweb.server.checkintegrity import reindex_entities
from cubicweb.dataimport.stores import NoHookRQLObjectStore
from cubicweb.dataimport.importer import (
    SimpleImportLog,
    cwuri2eid,
)

from cubicweb_skos import lcsv, sobjects as skos

# If you want to add a vocabulary here, add it to the end of the list if you
# want to create it in migration script by simply calling import_seda_schemes
# for proper interaction with saem which attempt to allocate reproducible ARK
# identifier to them.
LCSV_FILES = [
    (title, rtype, etype, join(abspath(dirname(__file__)), "migration", "data", fname))
    for title, rtype, etype, fname in (
        # schemes extracted from SEDA 2 XSD
        (
            "SEDA 2 : Actions",
            "seda_final_action",
            "SEDAStorageRule",
            "final_action_storage_code_type.csv",
        ),
        (
            "SEDA 2 : Unités de mesure",
            "seda_unit",
            (
                "SEDAWidth",
                "SEDAHeight",
                "SEDADepth",
                "SEDADiameter",
                "SEDALength",
                "SEDAThickness",
            ),
            "measurement_units_type.csv",
        ),
        (
            "SEDA 2 : Unités de poids",
            "seda_unit",
            "SEDAWeight",
            "measurement_weight_units_type.csv",
        ),
        (
            "SEDA : Types de mot-clé",
            "seda_keyword_type_to",
            (),
            "code_keyword_type.csv",
        ),
        ("SEDA 2 : Status légaux", "seda_legal_status_to", (), "legal_status.csv"),
        (
            "SEDA : Niveaux de description",
            "seda_description_level_to",
            (),
            "level_type.csv",
        ),
        # schemes extracted from SEDA 2 XSD, completed to support earlier SEDA versions
        (
            "SEDA : Sort final",
            "seda_final_action",
            "SEDAAppraisalRule",
            "final_action_appraisal_code_type.csv",
        ),
        # schemes extracted from earlier SEDA versions
        (
            "SEDA : Durée d'utilité administrative",
            "seda_rule",
            "SEDASeqAppraisalRuleRule",
            "dua.csv",
        ),
        (
            "SEDA : Codes de restriction d'accès",
            "seda_rule",
            "SEDASeqAccessRuleRule",
            "access_control.csv",
        ),
        (
            "SEDA : Règles de diffusion",
            "seda_rule",
            "SEDASeqDisseminationRuleRule",
            "dissemination.csv",
        ),
        ("SEDA : Types d'objets-données", "seda_type_to", (), "document_type_code.csv"),
        # other schemes
        ("Types MIME", "seda_mime_type_to", (), "mime_types.csv"),
        ("Types d'évènement", "seda_event_type_to", (), "event_types.csv"),
        (
            "Encodages (extraits du schéma UN/CEFACT)",
            "seda_encoding_to",
            (),
            "encodings.csv",
        ),
        ("Formats de fichier (PRONOM)", "seda_format_id_to", (), "file_formats.csv"),
        (
            "Niveau de classification (IGI 1300)",
            "seda_classification_level",
            (),
            "classification_levels.csv",
        ),
        ("Langues (ISO-639-3)", (), (), "languages.csv"),
        (
            "Algorithmes d'empreinte",
            "seda_algorithm",
            "SEDABinaryDataObject",
            "digest_algorithms.csv",
        ),
        ("Catégories de fichier", "file_category", (), "file_categories.csv"),
        (
            "Langues restreintes (compatible SEDA 0.2)",
            ("seda_language_to", "seda_description_language_to"),
            (),
            "languages-seda02.csv",
        ),
    )
]


def lcsv_import(cnx, store, fname, scheme_uri, **kwargs):
    """Actually import LCSV data file."""
    with open(fname, "rb") as stream:
        extentities = skos.lcsv_extentities(stream, scheme_uri, ";", "utf-8")
        import_log = SimpleImportLog(fname)
        skos.store_skos_extentities(
            cnx,
            store,
            extentities,
            import_log,
            raise_on_error=True,
            extid_as_cwuri=False,
            **kwargs,
        )


def lcsv_check(cnx, store, fname, scheme_uri, separator=";", **kwargs):
    """Simply check data file consistency."""
    counter = count()

    def uri_generator(val):
        return str(next(counter)) + val

    with open(join(dirname(__file__), "migration", "data", fname), "rb") as stream:
        lcsv2rdf = lcsv.LCSV2RDF(
            stream,
            separator,
            "utf-8",
            # XXX drop once skos is released
            uri_generator=uri_generator,
            uri_cls=str,
        )
        list(lcsv2rdf.triples())

        # also check there are the expected number of separator for each line
        stream.seek(0)
        expected_separators = stream.readline().decode("utf-8").count(separator)
        for i, line in enumerate(stream):
            line = line.decode("utf-8")
            if line.count(separator) != expected_separators:
                linenum = i + 2
                raise AssertionError(
                    "Got %s %s on line %s of %s, %s where expected"
                    % (
                        line.count(separator),
                        separator,
                        linenum,
                        fname,
                        expected_separators,
                    )
                )


def init_seda_scheme(cnx, title):
    """Create a scheme to hold SEDA concepts with the given title.

    Separated function to be monkey-patched if one need to customize the store (eg saem).
    """
    description = "edition 2009" if title.startswith("SEDA :") else None
    return cnx.create_entity("ConceptScheme", title=title, description=description)


def get_store(cnx):
    """Return the store to be used to import LCSV data files.

    Separated function to be monkey-patched if one needs to customize the store (eg saem).
    """
    if cnx.repo.system_source.dbdriver == "postgres":
        from cubicweb.dataimport.massive_store import MassiveObjectStore

        return MassiveObjectStore(cnx, eids_seq_range=1000)
    else:
        return NoHookRQLObjectStore(cnx)


def import_seda_schemes(cnx, lcsv_import=lcsv_import, lcsv_files=LCSV_FILES):
    """Import all LCSV data files defined in LCSV_FILES."""
    extid2eid = cwuri2eid(cnx, ("ConceptScheme", "Label"))
    # concepts and external URIs may come from any source
    extid2eid.update(cwuri2eid(cnx, ("Concept", "ExternalUri")))
    store = get_store(cnx)
    for title, rtypes, etypes, fname in lcsv_files:
        if not cnx.find("ConceptScheme", title=title):
            print("importing", title.encode("utf-8"))
            scheme = init_seda_scheme(cnx, title)
            extid2eid[scheme.cwuri] = scheme.eid
            lcsv_import(cnx, store, fname, scheme.cwuri, extid2eid=extid2eid)
            if not isinstance(rtypes, tuple):
                rtypes = (rtypes,)
            for rtype in rtypes:
                rtype_e = cnx.find("CWRType", name=rtype).one()
                scheme.cw_set(scheme_relation_type=rtype_e)
            if not isinstance(etypes, tuple):
                etypes = (etypes,)
            for etype in etypes:
                etype_e = cnx.find("CWEType", name=etype).one()
                scheme.cw_set(scheme_entity_type=etype_e)
            store.flush()
    store.commit()
    store.finish()
    if not isinstance(store, NoHookRQLObjectStore):
        # when using the massive store, we need explicit reindexation
        reindex_entities(cnx.repo.schema, cnx, etypes=["Concept", "ConceptScheme"])



from cubicweb_seda.xsd2yams import MULTIPLE_CHILDREN

add_relation_type("ordering")

parent_rtypes = [rtype for _, rtype in MULTIPLE_CHILDREN]

for rtype in parent_rtypes:
    for (peid,) in rql(f"Any P GROUPBY P WHERE X {rtype} P HAVING COUNT(X) > 1"):
        for idx, child in enumerate(
            rql(f"Any X WHERE X {rtype} P, P eid %(p)s", {"p": peid}).entities()
        ):
            child.cw_set(ordering=idx + 1)
    commit(ask_confirm=False)

rql("SET X ordering 1 WHERE X ordering NULL")

commit(ask_confirm=False)



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""Library of views use there and there in the interface"""

from cubicweb import tags, view, _
from cubicweb.predicates import match_kwargs, is_instance, score_entity
from cubicweb_web.view import EntityView
from cubicweb_web.views import baseviews, tabs

from ..xsd import un_camel_case
from ..xsd2uicfg import FIRST_LEVEL_ETYPES
from . import CONTENT_ETYPE, add_subobject_link, dropdown_button


def add_links_from_rtype_roles(entity, rtype_roles, extraurlparams):
    """Return a list of links to be inserted in dropdown button to add subobjects linked to `entity`
    through relations defined in `rtype_roles`, depending on user's permission to add any of them.

    `rtype_roles` is list of 2-uples `(rtype, role)`.
    """
    links = [
        add_subobject_link(entity, rtype, role, extraurlparams, klass="")
        for rtype, role in rtype_roles
    ]
    return [link for link in links if link]


def display_add_button(w, links, label):
    if links:
        # No links if user cannot add any relation.
        w(dropdown_button(label, links))
        w(tags.div(klass="clearfix"))


class XSDMetaEntityView(EntityView):
    __regid__ = "seda.xsdmeta"

    def entity_call(self, entity, skip_one_card=False, with_annotation=True):
        cardinality = getattr(entity, "user_cardinality", "1")
        if cardinality != "1" or not skip_one_card:
            self.w(f' <span class="cardinality">[{cardinality}]</span>')
        if with_annotation and getattr(entity, "user_annotation", None):
            description = entity.printable_value("user_annotation")
            if entity.cw_etype in FIRST_LEVEL_ETYPES:
                # skip first line, already displayed since it's used as title for the entity
                description = "\n".join(description.splitlines()[1:])
            if description:
                self.w(f' <span class="description text-muted">{description}</span>')


class TextEntityAttributeView(EntityView):
    """Attribute view for SEDA relations displaying entity as text (no link to the entity)"""

    __regid__ = "seda.reledit.text"
    subvid = "seda.business"

    def entity_call(self, entity, **kwargs):
        self.w("<div>")
        entity.view(self.subvid, w=self.w)
        entity.view("seda.xsdmeta", w=self.w)
        self.w("</div>")


class SimpleLinkToEntityAttributeView(TextEntityAttributeView):
    """Attribute view for SEDA relations linking directly to an entity without
    intermediary entity"""

    __regid__ = "seda.reledit.simplelink"
    subvid = "oneline"


class ComplexLinkEntityAttributeView(EntityView):
    """Attribute view for SEDA relations linking to an entity through an intermediary entity holding
    cardinalities / annotation
    """

    __regid__ = "seda.reledit.complexlink"

    def entity_call(self, entity):
        rtype = self.cw_extra_kwargs["rtype"].replace("_from", "_to")
        value_rset = entity.related(rtype)
        self.w("<div>")
        if value_rset:
            self.w('<span class="value">')
            self._cw.view("csv", value_rset, w=self.w)
            self.w("</span>")
        else:
            self.wdata(self._cw._("<no value specified>"))
        entity.view("seda.xsdmeta", w=self.w)
        self.w("</div>")


def alternative_values(entity, parent_rtype):
    """Display entities under the given alternative `entity`.

    Those are found by introspecting the schema, and skipping the relation from the alternative to
    its parent (`parent_rtype`).
    """
    req = entity._cw
    alternatives = []
    for rschema, _targets, role in entity.e_schema.relation_definitions():
        rtype = rschema.type
        if rtype.startswith("seda_") and rtype != parent_rtype:
            target_rset = entity.related(rtype, role)
            if target_rset:
                alternatives.append(req.view("seda.type_meta", rset=target_rset))
    return (f" <b>{req._(' ALT_I18N ')}</b> ").join(alternatives)


class AlternativeEntityAttributeView(EntityView):
    """Attribute view for SEDA alternative entities"""

    __regid__ = "seda.reledit.alternative"

    def entity_call(self, entity):
        alternatives = alternative_values(entity, self.cw_extra_kwargs["rtype"])
        self.w('<div class="alternative">')
        if alternatives:
            self.w(alternatives)
        else:
            self.wdata(self._cw._("<no value specified>"))
        entity.view("seda.xsdmeta", w=self.w)
        self.w("</div>")


class BusinessValueEntityView(EntityView):
    """Entity view that will display value of the attribute specified by `value_attr` on the
    entity's class
    """

    __regid__ = "seda.business"
    no_value_msg = _("<no value specified>")

    def entity_call(self, entity):
        value = self.entity_value(entity)
        if value:
            self.w('<span class="value">')
            self.w(value)
            self.w("</span>")
        else:
            self.wdata(self._cw._(self.no_value_msg))

    def entity_value(self, entity):
        if entity.value_attr:
            value = entity.printable_value(entity.value_attr)
        else:
            value = None
        return value


class BusinessValueSEDALegalStatustEntityView(BusinessValueEntityView):
    """Specific business value view for SEDALegalStatus entity: display the
    concept's label as text."""

    __select__ = is_instance("SEDALegalStatus")

    def entity_value(self, entity):
        if entity.seda_legal_status_to:
            return entity.seda_legal_status_to[0].view(self.__regid__)
        return super().entity_value(entity)


class BusinessValueConceptEntityView(BusinessValueEntityView):
    """Specific business value view for concept entity: simply display the
    concept's label as text."""

    __select__ = is_instance("Concept")

    def entity_value(self, entity):
        return entity.label()


class BusinessValueLinkEntityView(BusinessValueEntityView):
    """Similar to seda.business but value is enclosed in a link if some value is specified."""

    __abstract__ = True

    def entity_value(self, entity):
        value = super().entity_value(entity)
        if value:
            value = tags.a(value, href=entity.absolute_url())
        return value


class BusinessValueReferenceEntityView(BusinessValueEntityView):
    """Similar to seda.business but value is enclosed in a link if some value is specified."""

    __select__ = is_instance(*FIRST_LEVEL_ETYPES)

    def entity_value(self, entity):
        value = entity.dc_title()
        return tags.a(value, href=entity.absolute_url())


class TypeAndMetaEntityView(EntityView):
    """Glue entity's type, seda.business and seda.xsdmeta views together, for
    use within alternative"""

    __regid__ = "seda.type_meta"

    def entity_call(self, entity):
        self.w(entity.dc_type() + " ")
        entity.view("seda.business", w=self.w)
        entity.view("seda.xsdmeta", w=self.w, skip_one_card=True)


class ListItemView(EntityView):
    """Extended 'oneline' view for entities related to an AuthorityRecord, including link to remove
    the relation.
    """

    __regid__ = "seda.listitem"
    __select__ = EntityView.__select__ & match_kwargs("parent", "tabid")

    def entity_call(self, entity, parent, tabid, edit=True, delete=True):
        entity.view("seda.listitem.content", w=self.w)
        if entity.cw_has_perm("update"):
            self._cw.add_js(("cubicweb.ajax.js", "cubes.seda.js"))
            editurlparams = {
                "__redirectpath": parent.rest_path(),
                "__redirectparams": "tab=" + tabid,
            }
            if edit or delete:
                self.w('<div class="pull-right">')
                if edit:
                    self.w(
                        tags.a(
                            title=self._cw.__("edit"),
                            klass="icon-pencil",
                            href=entity.absolute_url(vid="edition", **editurlparams),
                        )
                    )
                if delete:
                    self.w(
                        tags.a(
                            title=self._cw.__("delete"),
                            klass="icon-trash",
                            href=entity.absolute_url(vid="deleteconf", **editurlparams),
                        )
                    )
                self.w("</div>")


class ListItemContentView(EntityView):
    """Glue seda.business and seda.xsdmeta views together, for use within list."""

    __regid__ = "seda.listitem.content"

    def entity_call(self, entity):
        entity.view("seda.business", w=self.w)
        entity.view("seda.xsdmeta", w=self.w, skip_one_card=True)


class PrimaryTabWithoutBoxes(tabs.PrimaryTab):
    """Abstract base class for tabs which rely on the primary view logic but
    don't want side boxes."""

    __abstract__ = True
    __regid__ = None  # we don't want 'primary'

    def is_primary(self):
        return False

    def content_navigation_components(self, context):
        pass


class SubObjectsTab(tabs.TabsMixin, EntityView):
    """Base class for tabs with a 'add' button and one or more list of subobjects, driven by the
    `rtype_role_targets` class attribute
    """

    rtype_role_targets = None
    subvid = "seda.listitem"

    @property
    def tabid(self):
        return self.__regid__

    def entity_call(self, entity):
        self.display_add_button(entity)
        self.display_subobjects_list(entity)

    def display_add_button(self, entity):
        rtype_roles = [(rtype, role) for rtype, role, _ in self.rtype_role_targets]
        links = add_links_from_rtype_roles(entity, rtype_roles, self.url_params(entity))
        display_add_button(self.w, links, self._cw.__("add"))

    def display_subobjects_list(self, entity):
        for rtype, role, target in self.rtype_role_targets:
            rset = entity.related(rtype, role)
            if rset:
                if target is not None:
                    self.w(f"<h2>{self._cw.__(target + '_plural')}</h2>")
                self._cw.view(
                    "list",
                    rset=rset,
                    parent=self.parent(entity),
                    w=self.w,
                    subvid=self.subvid,
                    tabid=self.tabid,
                )

    def url_params(self, entity):
        return {"__redirectparams": "tab=" + self.tabid}

    def parent(self, entity):
        return entity


# references views #################################################################################


def typed_reference_origin(entity):
    """Given a typed reference (eg. SEDAIsVersionOf, SEDAReplaces, etc), return a value to display
    its type and origin (archive unit defining it).
    """
    up_to_content_rtype = f"seda_{un_camel_case(entity.cw_etype[4:])}"
    content = entity.related(up_to_content_rtype).one()
    archive_unit = content.cw_adapt_to("ITreeBase").parent()
    return f"{archive_unit.view('incontext')} ({entity.dc_type()})"


class ArchiveUnitRefIdOutOfContextView(baseviews.OutOfContextView):
    """Custom outofcontext view for proper display of SEDAArchiveUnitRefId from in boxes"""

    __select__ = is_instance("SEDAArchiveUnitRefId")

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        parent = entity.cw_adapt_to("IContained").parent
        if parent.cw_etype == "SEDAAltArchiveUnitArchiveUnitRefId":
            parent.view("incontext", w=self.w)
        else:
            parent_parent_rtype = f"seda_{un_camel_case(parent.cw_etype[4:])}"
            parent_parent = parent.related(parent_parent_rtype, "object").one()
            self.w(typed_reference_origin(parent_parent))


class ArchiveUnitRefIdBusinessValueLinkEntityView(BusinessValueLinkEntityView):
    __select__ = is_instance("SEDAArchiveUnitRefId")

    def entity_value(self, entity):
        if entity.seda_archive_unit_ref_id_to:
            archiveunit = entity.seda_archive_unit_ref_id_to[0]
            return archiveunit.view("oneline")
        return None


def is_main_data_object_reference(reference):
    """Return True if the given SEDADataObjectReference is a 'main' reference from an archive unit
    to a data object (not a qualified reference, eg IsVersionOf, IsPartOf, Reference, etc.).
    """
    parent = reference.seda_data_object_reference[0]
    return parent.cw_etype == CONTENT_ETYPE


class MainDataObjectReferenceBusinessValueLinkEntityView(BusinessValueLinkEntityView):
    __select__ = is_instance("SEDADataObjectReference") & score_entity(
        lambda x: is_main_data_object_reference(x)
    )

    def entity_call(self, entity):
        do_rset = entity.related("seda_data_object_reference_id", "subject")
        if do_rset:
            do_rset.one().view(self.__regid__, w=self.w)
        else:
            super().entity_call(entity)


class OtherDataObjectReferenceBusinessValueLinkEntityView(BusinessValueLinkEntityView):
    __select__ = is_instance("SEDADataObjectReference") & score_entity(
        lambda x: not is_main_data_object_reference(x)
    )

    def entity_value(self, entity):
        if entity.seda_data_object_reference_id:
            dataobject = entity.seda_data_object_reference_id[0]
            return dataobject.view("oneline")
        return None


class ObjectReferencepArchiveUnitEntityView(view.EntityView):
    __regid__ = "seda.object-ref.archive-unit"
    __select__ = is_instance("SEDADataObjectReference")

    def entity_call(self, entity):
        au_seq = entity.seda_data_object_reference[0]
        if au_seq.cw_etype == CONTENT_ETYPE:
            au_choice = (
                au_seq.reverse_seda_seq_alt_archive_unit_archive_unit_ref_id_management[
                    0
                ]
            )
            target = au_choice.reverse_seda_alt_archive_unit_archive_unit_ref_id[0]
        else:
            rtype = f"seda_{un_camel_case(au_seq.cw_etype[4:])}"
            target = au_seq.related(rtype, "object").one()
        target.view("listitem", w=self.w)


class ReferenceListItemView(view.EntityView):
    __regid__ = "listitem"
    # additional match_kwargs because we don't want to be selected by SubObjectsTab based list
    __select__ = is_instance(
        "SEDAIsVersionOf",
        "SEDAReplaces",
        "SEDARequires",
        "SEDAIsPartOf",
        "SEDAReferences",
    ) & ~match_kwargs("tabid")

    def entity_call(self, entity):
        self.w(typed_reference_origin(entity))



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda views configuration / overrides for simplified profiles."""

from logilab.common.registry import objectify_predicate

from cubicweb.predicates import is_instance
from cubicweb_web.views import uicfg, autoform

from ..xsd2yams import RULE_TYPES
from ..entities import simplified_profile
from . import CONTENT_ETYPE, widgets

# ensure those are registered first
from . import mgmt_rules, archivetransfer, dataobject, archiveunit  # noqa


# Add arbitrary score using `yes` to overtake e.g. afs for DataObjectReference defined in content
afs = uicfg.autoform_section
simplified_afs = afs.derive(__name__, simplified_profile())
pvs = uicfg.primaryview_section
simplified_pvs = pvs.derive(__name__, simplified_profile())

# appraisal/access rules have a single top level cardinality in simplified profile, as well as
# always a start date. This implies:
# 1. force one and only one eg AppraisalRuleRule,
# 2. force start date, but hide it,
# 3. hide eg AppraisalRuleRule's cardinality.

# 3. hide rule rule's cardinality
simplified_afs.tag_attribute(
    ("SEDASeqAppraisalRuleRule", "user_cardinality"), "main", "hidden"
)
simplified_afs.tag_attribute(
    ("SEDASeqAccessRuleRule", "user_cardinality"), "main", "hidden"
)


# 2. hide start date's cardinality - uicfg is not enough since we want it when edited in the context
# of archive unit's content but not of rule entity types


class StartDateAutomaticEntityForm(autoform.AutomaticEntityForm):
    __select__ = is_instance("SEDAStartDate") & simplified_profile()

    def editable_attributes(self, strict=False):
        """return a list of (relation schema, role) to edit for the entity"""
        attributes = list(super().editable_attributes(strict))
        if self.linked_to.get(("seda_start_date", "subject")):
            eid = self.linked_to[("seda_start_date", "subject")][0]
            start_date_of = self._cw.entity_from_eid(eid).cw_etype
        elif self.edited_entity.has_eid():
            start_date_of = self.edited_entity.seda_start_date[0].cw_etype
        else:
            start_date_of = self.cw_extra_kwargs.get("petype")
        if start_date_of != CONTENT_ETYPE:
            attributes.remove(("user_cardinality", "subject"))
        return attributes


class RuleAutomaticEntityForm(autoform.AutomaticEntityForm):
    __select__ = (
        is_instance("SEDAAppraisalRule", "SEDAAccessRule") & simplified_profile()
    )

    def should_display_inline_creation_form(self, rschema, existing, card):
        # 1. force creation of one appraisal/access rule
        if not existing and rschema in (
            "seda_seq_appraisal_rule_rule",
            "seda_seq_access_rule_rule",
        ):
            return True
        return super().should_display_inline_creation_form(rschema, existing, card)

    def should_display_add_new_relation_link(self, rschema, existing, card):
        # 1. don't allow creation of more than one appraisal/access rule
        if rschema in ("seda_seq_appraisal_rule_rule", "seda_seq_access_rule_rule"):
            return False
        return super().should_display_add_new_relation_link(rschema, existing, card)


class SeqRuleRuleAutomaticEntityForm(autoform.AutomaticEntityForm):
    __select__ = (
        is_instance("SEDASeqAppraisalRuleRule", "SEDASeqAccessRuleRule")
        & simplified_profile()
    )

    def should_display_inline_creation_form(self, rschema, existing, card):
        # 2. force start date
        if not existing and rschema == "seda_start_date":
            return True
        return super().should_display_inline_creation_form(rschema, existing, card)


@objectify_predicate
def simplified_rule_rule(cls, req, rtype=None, pform=None, **kwargs):
    # check we're within a simplified profile
    if isinstance(pform, RuleAutomaticEntityForm) and rtype in (
        "seda_seq_appraisal_rule_rule",
        "seda_seq_access_rule_rule",
    ):
        return 1
    return 0


# 1. don't allow deletion of our appraisal/access rule


class RuleRuleInlineEntityEditionFormView(autoform.InlineEntityEditionFormView):
    __select__ = (
        autoform.InlineEntityEditionFormView.__select__ & simplified_rule_rule()
    )
    removejs = None


class RuleRuleInlineEntityCreationFormView(autoform.InlineEntityCreationFormView):
    __select__ = (
        autoform.InlineEntityCreationFormView.__select__ & simplified_rule_rule()
    )
    removejs = None


# 2. hide start date


@objectify_predicate
def simplified_start_date(cls, req, rtype=None, pform=None, **kwargs):
    # check we're within a simplified profile
    if isinstance(pform, SeqRuleRuleAutomaticEntityForm) and rtype == "seda_start_date":
        return 1
    return 0


class StartDateInlineEntityEditionFormView(autoform.InlineEntityEditionFormView):
    __select__ = (
        autoform.InlineEntityEditionFormView.__select__ & simplified_start_date()
    )
    removejs = None
    form_renderer_id = "notitle"


class StartDateInlineEntityCreationFormView(autoform.InlineEntityCreationFormView):
    __select__ = (
        autoform.InlineEntityCreationFormView.__select__ & simplified_start_date()
    )
    removejs = None
    form_renderer_id = "notitle"


# simplified profil will have a single appraisal/access rule, hence we can remove all the
# inheritance control mecanism: override will simply mean redefining a rule at some point in the
# tree.
for rule_type in RULE_TYPES:
    etype = f"SEDA{rule_type.capitalize()}Rule"
    rtype = f"seda_alt_{rule_type}_rule_prevent_inheritance"
    simplified_afs.tag_subject_of((etype, rtype, "*"), "main", "hidden")


# SEDAArchiveTransfer customization
for rtype, role in archivetransfer.at_ordered_fields:
    if not rtype.endswith("agency"):
        assert role == "object"
        if rtype in ("seda_archival_agreement", "seda_comment"):
            simplified_section = "attributes"
        else:
            simplified_section = "hidden"
        simplified_pvs.tag_object_of(
            ("*", rtype, "SEDAArchiveTransfer"), simplified_section
        )


# SEDAArchiveUnit
simplified_au_rtypes = {
    rtype
    for rtype, role, targets in archiveunit.SimplifiedContentMainView.rtype_role_targets
}
for rtype, role in archiveunit.content_ordered_fields:
    if role == "object":
        section = "inlined" if rtype in simplified_au_rtypes else "hidden"
        simplified_afs.tag_object_of(("*", rtype, CONTENT_ETYPE), "main", section)


# SEDABinaryDataObject/SEDAPhysicalDataObject customization
simplified_afs.tag_object_of(
    ("SEDADataObjectReference", "seda_data_object_reference_id", "*"), "main", "inlined"
)
simplified_afs.tag_object_of(
    ("SEDACustodialHistoryFile", "seda_data_object_reference_id", "*"), "main", "hidden"
)
simplified_pvs.tag_object_of(("*", "seda_data_object_reference_id", "*"), "hidden")
for rtype in ("seda_compressed", "seda_data_object_version_from"):
    simplified_pvs.tag_object_of(("*", rtype, "*"), "hidden")

simplified_pvs.tag_object_of(
    ("*", "seda_date_created_by_application", "SEDABinaryDataObject"), "attributes"
)


class DataObjectSimplifiedAutomaticEntityForm(widgets.SimplifiedAutomaticEntityForm):
    """On creation of a BinaryDataObject or PhysicalDataObject's in the context of a simplified
    profile, add a field to handle the creation of the relation to the archive unit specified in
    `req.form`.
    """

    # don't add match_form_params('referenced_by') since it's only specified for creation, not
    # edition
    __select__ = (
        widgets.SimplifiedAutomaticEntityForm.__select__
        & is_instance("SEDABinaryDataObject", "SEDAPhysicalDataObject")
        & simplified_profile()
    )

    def inlined_form_views(self):
        views = list(super().inlined_form_views())
        ref_forms = [
            v.form for v in views if v.rtype == "seda_data_object_reference_id"
        ]
        if ref_forms:  # may be empty if user has no write access
            ref_form = ref_forms[0]
            if not ref_form.edited_entity.has_eid() and not ref_form.posting:
                ref_form.add_hidden(
                    name="seda_data_object_reference",
                    eidparam=True,
                    role="subject",
                    value=self._cw.form["referenced_by"],
                )
        return views


# SEDADataObjectReference
simplified_afs.tag_subject_of(
    ("SEDADataObjectReference", "seda_data_object_reference_id", "*"),
    "inlined",
    "hidden",
)
simplified_afs.tag_attribute(
    ("SEDADataObjectReference", "user_cardinality"), "inlined", "hidden"
)


class DataObjectReferenceNoTitleEntityInlinedFormRenderer(
    widgets.NoTitleEntityInlinedFormRenderer
):
    """Don't display any title nor remove link for DataObjectReference in the context of a
    simplified profile.
    """

    __select__ = (
        widgets.NoTitleEntityInlinedFormRenderer.__select__
        & is_instance("SEDADataObjectReference")
        & simplified_profile()
    )



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda hooks"""

from collections import defaultdict
import itertools

from yams import ValidationError
from yams.schema import role_name

from cubicweb import _
from cubicweb.predicates import is_instance, score_entity
from cubicweb.server import hook

from .xsd2yams import MULTIPLE_CHILDREN
from .entities import rule_type_from_etype, diag
from .entities.itree import next_child_ordering, prepare_child_removal
from .entities.generated import (
    CHOICE_RTYPE,
    CHECK_CARD_ETYPES,
    CHECK_CHILDREN_CARD_RTYPES,
)


SEDA_PARENT_RTYPES = {}
CHOICE_RTYPE_ROLE = dict(itertools.chain(*CHOICE_RTYPE.values()))
CHOICE_RTYPES = set(CHOICE_RTYPE_ROLE)


class SetContainerOp(hook.DataOperationMixIn, hook.Operation):
    def precommit_event(self):
        cnx = self.cnx
        # build a mapping <eid>: <container eid>
        containers = {}
        data = dict(self.get_data())  # data is a 2-uple (eid, parent eid)
        for eid in data:
            # search for the container from the parent
            peid = data[eid]
            # climb up to the uppermost parent
            while peid in data:
                peid = data[peid]
            entity = cnx.entity_from_eid(peid)
            # we may run into caching issue if this operation is triggered several times in a
            # transaction
            entity.cw_clear_all_caches()
            adapter = entity.cw_adapt_to("IContained")
            if adapter is None:  # we reached the container
                containers[eid] = peid
            else:  # the adapter must know the container at this point
                if adapter.container is not None:
                    containers[eid] = adapter.container.eid
                else:
                    assert entity.cw_adapt_to(
                        "IContainer"
                    ), f"can't retrieve container for entity {entity}"
                    containers[eid] = peid
        # turn previous mapping into <container eid>: [<list of contained eids>]
        contained = defaultdict(list)
        for eid, ceid in containers.items():
            contained[ceid].append(eid)
        # and uses it to insert data using raw SQL for performance
        cursor = cnx.cnxset.cu
        for ceid, eids in contained.items():
            args = [{"x": eid} for eid in eids]
            cursor.executemany(
                "INSERT INTO container_relation(eid_from, eid_to)"
                " SELECT %%(x)s, %(c)s"
                " WHERE NOT EXISTS(SELECT 1 FROM container_relation"
                " WHERE eid_from=%%(x)s AND eid_to=%(c)s)" % {"c": ceid},
                args,
            )


class SetContainerHook(hook.Hook):
    __regid__ = "seda.graph.updated"
    __select__ = hook.Hook.__select__ & hook.match_rtype_sets(SEDA_PARENT_RTYPES)
    events = ("before_add_relation",)
    category = "metadata"

    def __call__(self):
        if SEDA_PARENT_RTYPES[self.rtype] == "subject":
            SetContainerOp.get_instance(self._cw).add_data((self.eidfrom, self.eidto))
        else:
            SetContainerOp.get_instance(self._cw).add_data((self.eidto, self.eidfrom))


class DeleteArchiveUnitContainerHook(hook.Hook):
    """Delete the "container" relation on SEDAArchiveUnit upon deletion of
    "seda_archive_unit" relation.

    Normally, deletion of "container" relation occurs as a consequence of the
    container entity deletion. But for SEDAArchiveUnit, the relation is
    optional so deleting the container entity does not delete the relation.
    """

    __regid__ = "seda.archive-unit.delete-container"
    __select__ = hook.Hook.__select__ & hook.match_rtype("seda_archive_unit")
    events = ("before_delete_relation",)

    def __call__(self):
        self._cw.execute("DELETE X container Y WHERE X eid %(x)s", {"x": self.eidfrom})


class CheckRefNonRuleIdCodeListHook(hook.Hook):
    """Watch for addition of concept through seda_ref_non_rule_id_to relation, to ensure it belongs
    to the scheme specified on the transfer

    This depends on the parent entity type and can not properly be handled with a rql constraint,
    since it would require several disjoint constraints, while cw's semantic is that all constraints
    should be matched.
    """

    __regid__ = "seda.schema.ref_non_rule_id"
    __select__ = hook.Hook.__select__ & hook.match_rtype("seda_ref_non_rule_id_to")
    events = ("after_add_relation",)

    def __call__(self):
        CheckRefNonRuleIdCodeListOp(self._cw, parent=self.eidfrom, concept=self.eidto)


# Late operation to be called once `container` relation is set
class CheckRefNonRuleIdCodeListOp(hook.LateOperation):
    def precommit_event(self):
        parent = self.cnx.entity_from_eid(self.parent, etype="SEDARefNonRuleId")
        # generate constraint on the concept's scheme depending on the parent's parent type
        parent_parent = parent.cw_adapt_to("IContained").parent
        rule_type = rule_type_from_etype(parent_parent.cw_etype)
        rql = (
            "Any C WHERE C eid %(c)s, C in_scheme CS, X eid %(x)s, X container AT, "
            "CACLV seda_{0}_rule_code_list_version_from AT, "
            "CACLV seda_{0}_rule_code_list_version_to CS".format(rule_type)
        )
        if not self.cnx.execute(rql, {"c": self.concept, "x": parent_parent.eid}):
            msg = _("this concept doesn't belong to scheme specified on the profile")
            raise ValidationError(
                parent.eid, {role_name("seda_ref_non_rule_id_to", "subject"): msg}
            )


class EnsureChoiceNotEmptyOp(hook.DataOperationMixIn, hook.Operation):
    """Make sure that, when a choice element is created/updated, it is not empty."""

    def precommit_event(self):
        for entity in self.get_data():
            if self.cnx.deleted_in_transaction(entity.eid):
                continue
            rtype_roles = CHOICE_RTYPE[entity.cw_etype]
            for rtype, role in rtype_roles:
                if entity.related(rtype, role=role):
                    break
            else:
                msg = _("An alternative cannot be empty")
                raise ValidationError(entity.eid, {"": msg})


class EnsureChoiceNotEmptyAtCreationHook(hook.Hook):
    """Make sure that, when a choice element is created, it is not empty."""

    __regid__ = "seda.choice.creation.ensure-not-empty"
    __select__ = hook.Hook.__select__ & score_entity(
        lambda x: x.cw_etype.startswith("SEDAAlt")
    )
    events = ("before_add_entity",)

    def __call__(self):
        EnsureChoiceNotEmptyOp.get_instance(self._cw).add_data(self.entity)


class EnsureChoiceNotEmptyAtRTypeDeletionHook(hook.Hook):
    """Make sure that, when a relation with a choice element is deleted, the choice remains not
    empty."""

    __regid__ = "seda.choice.relation-deletion.ensure-not-empty"
    __select__ = hook.Hook.__select__ & hook.match_rtype_sets(CHOICE_RTYPES)
    events = ("before_delete_relation",)

    def __call__(self):
        eid = {"object": self.eidto, "subject": self.eidfrom}[
            CHOICE_RTYPE_ROLE[self.rtype]
        ]
        choice = self._cw.entity_from_eid(eid)
        if choice.cw_etype in CHOICE_RTYPE:
            # Else this is an ambiguous relation, which in this case isn't
            # targetting a choice.
            EnsureChoiceNotEmptyOp.get_instance(self._cw).add_data(choice)


class SetDefaultCodeListVersionsHook(hook.Hook):
    """Hook triggering an operation to set sensible default values for a transfer's code list
    versions.
    """

    __regid__ = "seda.transfer.default-code-lists"
    __select__ = hook.Hook.__select__ & is_instance("SEDAArchiveTransfer")
    events = ("after_add_entity",)

    def __call__(self):
        SetDefaultCodeListVersionsOp(self._cw, transfer_eid=self.entity.eid)


class SetDefaultCodeListVersionsOp(hook.Operation):
    """Set sensible default values for a transfer's code list versions."""

    # XXX factorize with data structure in dataimport?
    complex_rtypes = [
        (
            "seda_message_digest_algorithm_code_list_version",
            "seda_algorithm",
            "SEDABinaryDataObject",
        ),
        ("seda_encoding_code_list_version", "seda_encoding_to", None),
        (
            "seda_data_object_version_code_list_version",
            "seda_data_object_version_to",
            None,
        ),
        ("seda_relationship_code_list_version", "seda_type_relationship", None),
        ("seda_access_rule_code_list_version", "seda_rule", "SEDASeqAccessRuleRule"),
        (
            "seda_appraisal_rule_code_list_version",
            "seda_rule",
            "SEDASeqAppraisalRuleRule",
        ),
        (
            "seda_dissemination_rule_code_list_version",
            "seda_rule",
            "SEDASeqDisseminationRuleRule",
        ),
        # 'seda_compression_algorithm_code_list_version',
        # 'seda_classification_rule_code_list_version',
        # 'seda_reuse_rule_code_list_version',
        # 'seda_storage_rule_code_list_version',
        # 'seda_acquisition_information_code_list_version',
        # instead of linking those two to vocabularies related to their
        # respective relation (seda_format_id_to and seda_mime_type_to), use the
        # file categories vocabulary for conveniency
        ("seda_file_format_code_list_version", "file_category", None),
        ("seda_mime_type_code_list_version", "file_category", None),
    ]

    def precommit_event(self):
        transfer = self.transfer_eid
        cnx = self.cnx
        for rtype, ref_rtype, ref_etype in self.complex_rtypes:
            etype = "SEDA" + "".join(word.capitalize() for word in rtype.split("_")[1:])
            rql = (
                "INSERT {etype} X: X {rtype}_from T, X {rtype}_to CS "
                "WHERE NOT Y {rtype}_from T, T eid %(t)s, "
                "CS scheme_relation_type RT, RT name %(rt)s"
            )
            if ref_etype is not None:
                rql += ", CS scheme_entity_type ET, ET name %(et)s"
            cnx.execute(
                rql.format(etype=etype, rtype=rtype),
                {"t": transfer, "rt": ref_rtype, "et": ref_etype},
            )


class SetWrapObjectHook(hook.Hook):
    """Ensure simplifed profiles has wrap_object relation set to False."""

    __regid__ = "seda.transfer.wrap_object"
    __select__ = hook.Hook.__select__ & is_instance("SEDAArchiveTransfer")
    events = ("after_add_entity", "after_update_entity")

    def __call__(self):
        if "simplified_profile" in self.entity.cw_edited:
            if (
                self.entity.cw_edited["simplified_profile"]
                and not self.entity.wrap_dataobjects
            ):
                self.entity.cw_set(wrap_dataobjects=False)


class SetDefaultDataObjectRefCardinalityHook(hook.Hook):
    """Hook triggering an operation to set cardinality to 1 on creation of a data object 'typed'
    reference.
    """

    __regid__ = "seda.doref.default"
    __select__ = hook.Hook.__select__ & is_instance("SEDADataObjectReference")
    events = ("after_add_entity",)

    def __call__(self):
        SetDefaultDataObjectRefCardinalityOp(self._cw, do_ref_eid=self.entity.eid)


class SetDefaultDataObjectRefCardinalityOp(hook.Operation):
    """Set cardinality to 1 on creation of a data object 'typed' reference."""

    def precommit_event(self):
        do_ref = self.cnx.entity_from_eid(self.do_ref_eid)
        parent = do_ref.seda_data_object_reference[0]
        if parent.cw_etype != "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement":
            do_ref.cw_set(user_cardinality="1")


class SimplifiedProfileSyncDORefCardOnCreateHook(hook.Hook):
    """Hook triggering an operation to keep in sync user_cardinality of data object reference and
    its associated data object for simplified profiles, on creation of the relation between them.
    """

    __regid__ = "seda.doref.default"
    __select__ = hook.Hook.__select__ & hook.match_rtype(
        "seda_data_object_reference_id"
    )
    events = ("after_add_relation",)

    def __call__(self):
        ref = self._cw.entity_from_eid(self.eidfrom)
        bdo = self._cw.entity_from_eid(self.eidto)
        if (
            ref.cw_etype == "SEDADataObjectReference"
            and bdo.cw_etype == "SEDABinaryDataObject"
        ):
            SimplifiedProfileSyncDORefCardOp(self._cw, do_ref=ref, do=bdo)


class SimplifiedProfileSyncDORefCardOnUpdateHook(hook.Hook):
    """Hook triggering an operation to keep in sync user_cardinality of data object reference and
    its associated data object for simplified profiles, on update of the data object's cardinality.
    """

    __regid__ = "seda.doref.default"
    __select__ = hook.Hook.__select__ & is_instance("SEDABinaryDataObject")
    events = ("after_update_entity",)

    def __call__(self):
        if "user_cardinality" in self.entity.cw_edited:
            references = self.entity.reverse_seda_data_object_reference_id
            if len(references) == 1:  # else it can't be a simplified profile
                SimplifiedProfileSyncDORefCardOp(
                    self._cw, do_ref=references[0], do=self.entity
                )


class SimplifiedProfileSyncDORefCardOp(hook.Operation):
    """Keep in sync user_cardinality of data object reference and its associated data object for
    simplified profiles.
    """

    def precommit_event(self):
        container = (
            self.do.seda_binary_data_object[0]
            if self.do.seda_binary_data_object
            else None
        )
        if container and container.simplified_profile:
            self.do_ref.cw_set(user_cardinality=self.do.user_cardinality)


class DeleleBDOWhenDeletingDORefHook(hook.Hook):
    """Hook triggering an operation to delete the data object associated to a
    reference when the later is deleted from a simplified profile or an archive unit.
    """

    __regid__ = "seda.doref.del"
    __select__ = hook.Hook.__select__ & hook.match_rtype(
        "seda_data_object_reference_id"
    )
    events = ("after_delete_relation",)

    def __call__(self):
        ref = self._cw.entity_from_eid(self.eidfrom)
        bdo = self._cw.entity_from_eid(self.eidto)
        if (
            ref.cw_etype == "SEDADataObjectReference"
            and bdo.cw_etype == "SEDABinaryDataObject"
        ):
            if bdo.seda_binary_data_object:
                container = bdo.seda_binary_data_object[0]
                if not container.simplified_profile:
                    return
            else:
                if not bdo.container or bdo.container[0].cw_etype != "SEDAArchiveUnit":
                    return
            DeleleBDOAlongWithDORefOp(self._cw, do_ref=ref, do=bdo)


class DeleleBDOAlongWithDORefOp(hook.Operation):
    """Delete data object whose former reference has been deleted in the
    transaction. Expected to be run only if the profile is a simplified
    profile or if the data object is related to an archive unit not itself
    related to profile.
    """

    def precommit_event(self):
        if self.cnx.deleted_in_transaction(self.do_ref.eid):
            self.do.cw_delete()


class CheckProfileSEDACompatiblityOp(hook.DataOperationMixIn, hook.LateOperation):
    """Data operation that will check compatibility of a SEDA profile upon modification.

    This is a late operation since it has to be executed once the 'container' relation is set.
    """

    def precommit_event(self):
        cnx = self.cnx
        profiles = set()
        for eid in self.get_data():
            if cnx.deleted_in_transaction(eid):
                continue
            entity = cnx.entity_from_eid(eid)
            if entity.cw_etype == "SEDAArchiveTransfer":
                profiles.add(entity)
            else:
                container = entity.cw_adapt_to("IContained").container
                if (
                    container is not None
                    and container.cw_etype == "SEDAArchiveTransfer"
                ):
                    profiles.add(container)
        with cnx.deny_all_hooks_but():
            for profile in profiles:
                adapter = profile.cw_adapt_to("ISEDACompatAnalyzer")
                supported_formats = adapter.diagnose()
                profile.cw_set(compat_list=", ".join(sorted(supported_formats)))
                if profile.simplified_profile and "simplified" not in supported_formats:
                    raise ValidationError(
                        profile.eid,
                        {
                            "": _(
                                "This profile can't be turned to simplified. "
                                "See the diagnostic tab for more information"
                            )
                        },
                    )

    def add_entity(self, entity):
        """Add entity, its parent entities (up to the container root) for update of their
        modification date at commit time.
        """
        self.add_data(entity.eid)
        while True:
            safety_belt = {entity.eid}
            contained = entity.cw_adapt_to("IContained")
            if contained is None:
                assert entity.cw_adapt_to("IContainer")
                break
            else:
                entity = contained.container
                if entity is None or entity.eid in safety_belt:
                    break
                self.add_data(entity.eid)
                safety_belt.add(entity.eid)


class CheckNewProfile(hook.Hook):
    """Instantiate operation checking for profile seda compat on its creation"""

    __regid__ = "seda.transfer.created.checkcompat"
    __select__ = hook.Hook.__select__ & is_instance("SEDAArchiveTransfer")
    events = ("after_add_entity",)
    category = "metadata"

    def __call__(self):
        CheckProfileSEDACompatiblityOp.get_instance(self._cw).add_entity(self.entity)


WATCH_RTYPES_SET = set().union(
    *(
        (
            rtype
            for rtype in rule.watch
            # filter out (entity type, attribute)
            if not isinstance(rtype, tuple)
        )
        for rule in diag.RULES.values()
    )
)
WATCH_RTYPES_SET |= set(CHECK_CHILDREN_CARD_RTYPES)


class AddOrRemoveChildrenHook(hook.Hook):
    """Some relation involved in diagnosis of the profile is added or removed."""

    __regid__ = "seda.transfer.relupdated.checkcompat"
    __select__ = hook.Hook.__select__ & hook.match_rtype_sets(WATCH_RTYPES_SET)
    events = ("before_add_relation", "before_delete_relation")
    category = "metadata"

    def __call__(self):
        op = CheckProfileSEDACompatiblityOp.get_instance(self._cw)
        op.add_entity(self._cw.entity_from_eid(self.eidfrom))
        op.add_entity(self._cw.entity_from_eid(self.eidto))


WATCH_ETYPES = defaultdict(set)
for rule in diag.RULES.values():
    for etype_attr in rule.watch:
        if isinstance(etype_attr, tuple):
            WATCH_ETYPES[etype_attr[0]].add(etype_attr[1])

WATCH_ETYPES["SEDAArchiveTransfer"].add("simplified_profile")

for etype in CHECK_CARD_ETYPES:
    WATCH_ETYPES[etype].add("user_cardinality")


class UpdateWatchedProfileElementHook(hook.Hook):
    """Some entity involved in diagnosis of the profile is created or updated."""

    __regid__ = "seda.transfer.updated.checkcompat"
    __select__ = hook.Hook.__select__ & is_instance(*WATCH_ETYPES)
    events = ("before_add_entity", "before_update_entity")
    category = "metadata"

    def __call__(self):
        if WATCH_ETYPES[self.entity.cw_etype] & set(self.entity.cw_edited):
            CheckProfileSEDACompatiblityOp.get_instance(self._cw).add_entity(
                self.entity
            )


# XXX rather kill SEDAMimeType and SEDAFormatId if client is fine with "auto"
# cardinality handling
class InitBDO(hook.Hook):
    """On binary data object creation, link it to a SEDAMimeType and SEDAFormatId
    sub-entities.
    """

    __regid__ = "seda.ux.autoformat"
    __select__ = hook.Hook.__select__ & is_instance("SEDABinaryDataObject")
    events = ("after_add_entity",)

    def __call__(self):
        self._cw.create_entity(
            "SEDAMimeType", user_cardinality="0..1", seda_mime_type_from=self.entity
        )
        self._cw.create_entity(
            "SEDAFormatId", user_cardinality="0..1", seda_format_id_from=self.entity
        )


class SyncFileCategoryOp(hook.DataOperationMixIn, hook.LateOperation):
    """On file category change, synchronize associated mime types and format ids
    to the data object.

    This is a late operation because we have to go back to the container to know
    if it's a component archive unit or a transfer, so it has to be executed
    once the 'container' relation is set.
    """

    # set possible mime types / format ids by joining related transfer's
    # vocabularies to file category vocabulary using the concept's label
    set_mime_type_rql = (
        "SET X seda_mime_type_to MT WHERE X eid %(x)s, "
        "MT in_scheme CS, CACLV seda_mime_type_code_list_version_from AT, "
        "CACLV seda_mime_type_code_list_version_to CS, AT eid %(c)s, "
        "MT preferred_label MTL, MTL label MTLL, "
        "FCMT preferred_label FCMTL , FCMTL label MTLL, "
        # case where a file extension is given...
        "EXISTS(FCMT broader_concept EXT1, EXT1 eid IN ({eids}), "
        #   ... ensured by checking its parent is a root category
        "       EXISTS(EXT1 broader_concept CAT1, NOT EXISTS(CAT1 broader_concept Z1))) "
        # case where a generic category is given
        "OR EXISTS(FCMT broader_concept EXT2, EXT2 broader_concept CAT2, "
        "          CAT2 eid IN ({eids}), "
        #   ... ensured by checking it's a root category
        "          NOT EXISTS(CAT2 broader_concept Z2))"
    )
    set_format_id_rql = (
        "SET X seda_format_id_to FI WHERE X eid %(x)s, "
        "FI in_scheme CS, CACLV seda_file_format_code_list_version_from AT, "
        "CACLV seda_file_format_code_list_version_to CS, AT eid %(c)s, "
        "FI preferred_label FIL, FIL label FILL, "
        "FCFI preferred_label FCFIL , FCFIL label FILL, "
        # case where a file extension is given...
        "EXISTS(FCFI broader_concept MT, MT broader_concept EXT1, EXT1 eid IN ({eids}), "
        #   ... ensured by checking its parent is a root category
        "       EXISTS(EXT1 broader_concept CAT1, NOT EXISTS(CAT1 broader_concept Z1))) "
        "OR EXISTS(FCFI broader_concept MT2, MT2 broader_concept EXT2, EXT2 broader_concept CAT2, "
        "          CAT2 eid IN ({eids}),"
        #   ... ensured by checking it's a root category
        "          NOT EXISTS(CAT2 broader_concept Z2))"
    )

    def precommit_event(self):
        for bdo_eid in self.get_data():
            if self.cnx.deleted_in_transaction(bdo_eid):
                continue
            bdo = self.cnx.entity_from_eid(bdo_eid)
            if not bdo.container:
                # we may have to clear the container cache if the parent has
                # been added in the same transaction
                bdo.cw_clear_relation_cache("container", "subject")

            container = bdo.cw_adapt_to("IContained").container
            if container.cw_etype != "SEDAArchiveTransfer":
                # don't afford doing this in components which are not bound to
                # vocabularies, it will be done upon import in a transfer
                continue
            bdo.mime_type.cw_set(seda_mime_type_to=None)
            bdo.format_id.cw_set(seda_format_id_to=None)
            if bdo.file_category:
                eids = ",".join(str(x.eid) for x in bdo.file_category)
                self.cnx.execute(
                    self.set_mime_type_rql.format(eids=eids),
                    {"x": bdo.mime_type.eid, "c": container.eid},
                )
                self.cnx.execute(
                    self.set_format_id_rql.format(eids=eids),
                    {"x": bdo.format_id.eid, "c": container.eid},
                )


class SyncFileCategoryHook(hook.Hook):
    """On file category change, instantiate an operation to synchronize
    associated mime types and format ids to the data object.
    """

    __regid__ = "seda.ux.filecategory"
    __select__ = hook.Hook.__select__ & hook.match_rtype("file_category")
    events = ("after_add_relation", "after_delete_relation")
    category = "metadata"

    def __call__(self):
        SyncFileCategoryOp.get_instance(self._cw).add_data(self.eidfrom)


class SetOrderingHook(hook.Hook):
    """Hook to set ordering attribute of child object when some relation is added."""

    __regid__ = "seda.ordering.set"
    __select__ = hook.Hook.__select__ & hook.match_rtype_sets(
        {rtype for _, rtype in MULTIPLE_CHILDREN}
    )

    events = ("after_add_relation",)

    def __call__(self):
        eid, parent_eid = self.eidfrom, self.eidto
        ordering = next_child_ordering(self._cw, parent_eid, self.rtype)
        entity = self._cw.entity_from_eid(eid)
        entity.cw_set(ordering=ordering)


class KeepOrderingSequentielHook(hook.Hook):
    """Hook to keep `ordering` attribute of sibbling entities sequential when some
    relation will be removed.
    """

    __regid__ = "seda.ordering.remove"
    __select__ = hook.Hook.__select__ & hook.match_rtype_sets(
        {rtype for _, rtype in MULTIPLE_CHILDREN}
    )

    events = ("before_delete_relation",)

    def __call__(self):
        prepare_child_removal(self._cw.entity_from_eid(self.eidfrom))


def registration_callback(vreg):
    from cubicweb.server import ON_COMMIT_ADD_RELATIONS
    from cubicweb_seda import seda_profile_container_def, iter_all_rdefs

    vreg.register_all(globals().values(), __name__)

    for etype, parent_rdefs in seda_profile_container_def(vreg.schema):
        for rtype, role in parent_rdefs:
            if SEDA_PARENT_RTYPES.setdefault(rtype, role) != role:
                raise Exception("inconsistent relation", rtype, role)
    # manually add seda_data_object_reference_id since it's not composite and as
    # such not considered as part of the graph, but it has to be watched by hook
    # triggering the operation to set the container for the case of component
    # archive unit
    SEDA_PARENT_RTYPES["seda_data_object_reference_id"] = "object"
    for rdef, role in iter_all_rdefs(vreg.schema, "SEDAArchiveTransfer"):
        ON_COMMIT_ADD_RELATIONS.add(str(rdef.rtype))



for e in rql("Any X groupby X WHERE X container C HAVING COUNT(C) > 1").entities():
    container = max(c.eid for c in e.container)
    e.cw_set(container=None)
    e.cw_set(container=container)
commit()

sync_schema_props_perms("container")

scheme = cnx.find("ConceptScheme", title="SEDA 2 : Status légaux").one()
for old_label, new_label in [
    ("Archive publique", "Archives publiques"),
    ("Archive privée", "Archives privées"),
]:
    rql(
        "SET X label %(new_label)s WHERE X label %(old_label)s, "
        "X label_of C, C in_scheme S, S eid %(s)s",
        {"s": scheme.eid, "new_label": new_label, "old_label": old_label},
    )

commit()

scheme = cnx.find("ConceptScheme", title="SEDA : Niveaux de description").one()
for old_label, new_label in [
    ("Dossier l'intérieur d'une série organique", "Dossier"),
    ("Item", "Pièce"),
]:
    rql(
        "SET X label %(new_label)s WHERE X label %(old_label)s, "
        "X label_of C, C in_scheme S, S eid %(s)s",
        {"s": scheme.eid, "new_label": new_label, "old_label": old_label},
    )
    # don't care about migrating updated definition for now
commit()


sync_schema_props_perms("seda_description_level")
